/* Michael J. Chappell  February 6, 1993  FREDKIN.C

   Demonstrates four-fold reproduction using a system described by Edward
   Fredkin of MIT.  If you add the number of cells above, below, to the
   left and to the right of a cell, and the sum is odd then the cell will
   be filled, otherwise, the cell will be empty.

   Version 2.0 using ncurses - December 8, 2021
   Version 2.1 - December 19, 2021
   - Added load, save, pause, help, verbose, modify speed
   Version 2.2 - December 21, 2021
   - Various bug fixes
*/

/* #include <bios.h> */
/* #include <conio.h> */

#include <ncurses/ncurses.h>
#include <unistd.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#define VERSION_NUMBER "2.2"
#define COPYRIGHT "Copyright 2021     Michael J. Chappell"
#define DEFAULT_DELAY 50000
/* 
   DEFAULT_FILENAME could be customized for the OS.  Avoid path components 
   here to simplify Windows and *NIX compatability.
*/
#define DEFAULT_FILENAME ""
/* 
   MAX_PATH_LENGTH is being set for the buffer and should be larger
   than what the OS allows.  I'm trying to reserve memory here to avoid
   using malloc and prevent problems from excessively long command lines.
*/
#define MAX_PATH_LENGTH 2048 

int buffer[2][27][82];

/* 
   Initialize(), Load_File() and Save_File() should all return 0
   Any other result indicates a file I/O problem we should terminate.
*/
int Initialize(int *generation, int *udelay, int argc, char *datafile);
int Load_File(int *generation, int *udelay, char *datafile);
int Save_File(int generation, int udelay, char *datafile);

void Help(int generation, int udelay, int argc, char *datafile);

void Buffer_To_Screen(int s);
void Redraw_Buffer(int s, int d);

int  main(int argc, char **argv) {
     int board=0, generation=1;
     int c;
     int udelay;
     char datafile[MAX_PATH_LENGTH];
     int verbose=FALSE;
     int done=FALSE;

     initscr();

     if (argc>1){
       strncpy(datafile, argv[1], MAX_PATH_LENGTH);
     } else {
       strncpy(datafile, DEFAULT_FILENAME, MAX_PATH_LENGTH);
     }
     
     if (Initialize(&generation, &udelay, argc, datafile)) {
       /* if initialize fails we will abort */
       addstr("Press any key to continue.\n");
       refresh();
     } else {
       nodelay(stdscr,TRUE);
       noecho();

       do {
	 Buffer_To_Screen(board);

	 move(0,0);
	 if (verbose) {
	 printw("Generation: %7d Delay: %7d File: %s",
		generation++, udelay, datafile);
	 }
	 refresh();

	 Redraw_Buffer(board, (board+1)&1);
	 board = (board+1)&1;
	 c = getch();
	 switch (tolower(c)) {
	 case '0':
	   udelay = DEFAULT_DELAY;
	   break;
	 case '8':
	   udelay = 1000000;
	   break;
	 case '7':
	   udelay = 100000;
	   break;
	 case '6':
	   udelay = 10000;
	   break;
	 case '5':
	   udelay = 1000;
	   break;
	 case '4':
	   udelay = 100;
	   break;
	 case '3':
	   udelay = 10;
	   break;
	 case '2':
	   udelay = 1;
	   break;
	 case '1':
	   udelay = 0;
	   break;

	 case 'd':
	   nodelay(stdscr,FALSE);
	   echo();
	   clear();
	   printw("Enter new delay.  \n");
	   refresh();
	   scanw("%d", &udelay);
	   nodelay(stdscr,TRUE);
	   noecho();
	   break;
	 case 'h':
	 case '?':
	 case '\027':
	   Help(generation, udelay, argc, datafile);
	   break;
	 case 'i':
	   strncpy(datafile, DEFAULT_FILENAME, MAX_PATH_LENGTH);
	   if (Initialize(&generation, &udelay, 1, datafile)) done=TRUE;
	   break;
	 case 'l':
	   nodelay(stdscr,FALSE);
	   echo();
	   clear();
	   printw("Load data from:  \n");
	   refresh();
	   scanw("%s", datafile);
	   noecho();
	   nodelay(stdscr,TRUE);
	   if (Load_File(&generation, &udelay, datafile)) done=TRUE;
	   break;
	 case 'p':
	 case ' ':
	   refresh();
	   echo();
	   nodelay(stdscr,FALSE);
	   getch();
	   refresh();
	   nodelay(stdscr,TRUE);
	   noecho();
	   break;
	 case 's':
	   nodelay(stdscr,FALSE);
	   echo();
	   clear();
	   printw("Save data to:  \n");
	   refresh();
	   scanw("%s", datafile);
	   nodelay(stdscr,TRUE);
	   noecho();
	   if (Save_File(generation, udelay, datafile)) done=TRUE;
	   break;
	 case 'v':
	   verbose = !verbose;
	   break;
	 case 'q':
	   done = TRUE;
	   break;
	   
	 case '-':
	   udelay -= 100;
	   break;
	 case '+':
	   udelay += 100;
	   break;
	 case '<':
	   udelay -= 1000;
	   break;
	 case '>':
	   udelay += 1000;
	   break;
	 case '[':
	   udelay -= 10000;
	   break;
	 case ']':
	   udelay += 10000;
	   break;
	 case '{':
	   udelay -= 100000;
	   break;
	 case '}':
	   udelay += 100000;
	   break;
	 }
	 udelay=(udelay>0)?udelay:0;
	 udelay=(udelay>1000000)?1000000:udelay;
	 
	 usleep(udelay);
       } while (!done);
     }
     echo();
     nodelay(stdscr,FALSE);
     getch();
     endwin();

     return EXIT_SUCCESS;
} /* main */

int Initialize(int *generation, int *udelay, int argc, char *datafile) {
     int i, j, k;

     *udelay = DEFAULT_DELAY;
     *generation = 1;
     
     if (argc==1) {
       for (k=0; k<2; k++)
	 for (j=0; j<27; j++)
           for (i=0; i<82; i++) buffer[k][j][i]=0;
       nodelay(stdscr,FALSE);
       echo();
       clear();
       do {
	 addstr("\nEnter 0 0 to end input.  Otherwise, the specified ");
	 addstr("\ncoordinates will be set if in range. [1,80] [1,25]  ");
	 refresh();
	 scanw("%d %d", &i, &j);
	 if ((0<i)&&(i<81)&&(0<j)&&(j<26)) buffer[0][j][i]=1;
       } while (i!=0);
       nodelay(stdscr,TRUE);
       noecho();
     } else {
       return Load_File(generation, udelay, datafile);
     }
     return 0;
} /* Initialize */

int Load_File(int *generation, int *udelay, char *datafile) {
  FILE *fp;
  char line[MAX_PATH_LENGTH];
  int i, j, k;

  for (k=0; k<2; k++)
    for (j=0; j<27; j++)
      for (i=0; i<82; i++) buffer[k][j][i]=0;
  /* 
     Ugly hack - throw away line 1
     generation and udelay should be on the next 2 lines
     We will assume a single string before the number and ignore it.
     Error checking needs to be implemented.
  */
  if (fp=fopen(datafile,"r")) {
    fscanf(fp, "%[^\n]", line);
    fscanf(fp, "%s %d", line, generation);
    fscanf(fp, "%s %d", line, udelay);
    do {
      fscanf(fp, "%d %d", &i, &j);
      if ((0<i)&&(i<81)&&(0<j)&&(j<26)) buffer[0][j][i]=1;
    } while (i!=0);
    fclose(fp);
  } else {
    fprintf(stderr, "Unable to read from file: %s\n", datafile);
    return -1;
  }
  return 0;
}

int Save_File(int generation, int udelay, char *datafile) {
  FILE *fp;
  int col, row;
  if (fp=fopen(datafile, "w")) {
    fprintf(fp, "# FREDKIN INPUT DATA MJC VERSION %s\n", VERSION_NUMBER);
    fprintf(fp, "Generation: %d\n", generation);
    fprintf(fp, "Delay: %d\n", udelay);
    for (row=1; row<26; row++) {
      for (col=1; col<81; col++) {
	if (buffer[0][row][col]) {
	  fprintf(fp, "%d %d\n", col, row);
	}
      }
    }
    fprintf(fp, "0 0\n");
    fclose(fp);
    return 0;
  } else {
    fprintf(stderr, "Unable to write to file: %s\n", datafile);
    return -1;
  }
}

void Help(int generation, int udelay, int argc, char *datafile) {
  char directory[MAX_PATH_LENGTH];

  getcwd(directory, MAX_PATH_LENGTH-1);
  clear();
  printw("Fredkin Version %s\n", VERSION_NUMBER);
  printw("%s\n\n", COPYRIGHT);
  if (argc==2) {
    printw("File              : %s\n", datafile);
  } else {
    printw("\n");
  }
  printw("Current directory : %s\n", directory);
  printw("Default uDelay    : %d\n", DEFAULT_DELAY);
  printw("Generation        : %d\n", generation);
  printw("uDelay            : %d\n\n", udelay);
  printw("Keys:\n"
	 "\t  D) Set uDelay."
	 "\t  P) Pause.\n"
	 "\t  H) Help.      "
	 "\t  Q) Quit.\n"
	 "\t  I) Initialize."
	 "\t  S) Save.\n"
	 "\t  L) Load.      "
	 "\t  V) Toggle Verbose\n\n"
	 "Keys to set uDelay:\n"
	 "\t  0) Default uDelay\n"
	 "\t  1)        0\t\t  5)     1,000\n"
	 "\t  2)        1\t\t  6)    10,000\n"
	 "\t  3)       10\t\t  7)   100,000\n"
	 "\t  4)      100\t\t  8) 1,000,000\n\n"
	 "\t-/+) Decrement/Increment uDelay by     100.\n"
	 "\t</>) Decrement/Increment uDelay by   1,000.\n"
	 "\t[/]) Decrement/Increment uDelay by  10,000.\n"
	 "\t{/}) Decrement/Increment uDelay by 100,000.\n"
	 );
  refresh();
  echo();
  nodelay(stdscr,FALSE);
  getch();
  refresh();
  nodelay(stdscr,TRUE);
  noecho();

  return;
}

void Redraw_Buffer(int s, int d) {
     int i, j, temp;

     for (i=1; i<26; i++)
       for (j=1; j<81; j++) {
           temp = buffer[s][i][j-1]+buffer[s][i][j+1]+buffer[s][i-1][j];
           temp += buffer[s][i+1][j];
           buffer[d][i][j] = temp&1;
       } /* for j,i */
} /* Redraw_Buffer */

void Buffer_To_Screen(int s) {
     int i, j;

     clear();
     refresh();
     for (j=1; j<25; j++) {
       for (i=1; i<80; i++) {
           if ((j<25)|(i<80)) addch(buffer[s][j][i]?'@':' ');
       }
       addstr("\n");
     } /* for i,j */
     refresh();
} /* Buffer_To_Screen */
